/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/*
   FTP client.
   Ping and traceroute (normal part of the stack).
   No daemons listening.
*/

/* standard includes */
#include <stdlib.h> 
#include <stdio.h>
#include <string.h>

/* fixups for HI-TECH C, GCC, and probably other compilers. */
#ifdef HI_TECH_C
#include <sys.h>
#include <unixio.h>
#include <conio.h>
#include <stat.h>
/* #include <cpm.h> */
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef LINUX
#include <asm/io.h>
#endif
#endif

#include "in.h"
#include "netinet.h"
#include "socket.h"
#include <ctype.h>

/* various fixups, os dependent */
#include "compat.h"
#include "netutil.h"

#define PROGRAM "CPMFTP"
#define VERSION "0.0.3"
#define EXECUTABLE "FTP"

int sock = -1; /* command socket */
int outsock = -1; /* outbound data socket */
int insock = -1; /* inbound data socket */
int crseen = 0;
int servstate = -1;
int port = 21;
unsigned int passport = 0;
int nopassport = 20; /* listen */
int was = 0; /* what command group */
int is = 0; /* the actual reply code */
int zap = -1;
int lpointer = 0;
int dav = 0;	/* server data line ready */
int xferfile; /* i/o file fd */
int sizecount = 0;
int hash = 0; /* hash prints on 128 bytes... */
int outx = 0;
int loggedin = 0;
int key = 0;
int depth = 0;
char *keyboard; /* buffered keystrokes */
char *cmdq; /* what the user typed */
char *iodata; /* what the server typed */
char *smackit; /* line of what the server typed */
char *xferdata; /* data to/from server, one sector */
char *cmdo; /* data TO the server */
unsigned char destination[4];
FILE *xyz;

#define BIGGEST 320

void doio(VOIDFIX)
{
	int i, j, k;
	/* inbound command socket */
	if((sock > 0)) {
		zap = sockread(sock, iodata + lpointer, (BIGGEST - 1) - lpointer);
		if(zap < 0 && lpointer == 0) {
			printf("\nread socket died with error %i\n", zap);
			sock = -1;
			exit(1);
		}
		if(zap > 0) lpointer += zap;
	}
	if((lpointer > 0) && (dav == 0)) {
		for (i = 0; i < lpointer; i++) {
			if((iodata[i] == 0x0d) || (iodata[i+1] == 0x0a)) {
				memset(smackit, 0, BIGGEST);
				memmove(smackit, iodata, i); /* copy string */
				smackit[i] = 0x00;
				dav++;
				i++;
				i++;
				lpointer = lpointer - i;
				memmove(iodata, iodata + i, lpointer);
				return; /* bail out here */
			}
		}
	}
	/* NOTE: once read() and write() are unified, this can become one routine */
	/* inbound data */
	while(insock > 0) {
		i=sockread(insock, xferdata, 128);
		if(i < 0) {
			sockclose(insock);
			insock = -1;
			return;
		}
		if((i > 0) && (xferfile > -1)) {
			j = 0;
			while(j != i) {
				k=fwrite(xferdata + j, 1, i - j, xyz);
				if(ferror(xyz) != 0) {
					sockclose(insock);
					insock = -1;
					return; /* let the caller close the file */
				}
				j = j + k;
			}
			if(hash != 0) {
				sizecount += i;
				while(sizecount > 127) {
					sizecount -= 128;
					putchar('#');
					fflush(stdout);
				}
			}
		}
	}
	/* outbound data */
	while(outsock > 0) {
		i = fread(xferdata, 1, 128, xyz);
		if(i == 0) {
			sockclose(outsock);
			outsock = -1;
			return;
		}
		if(xferfile > -1) {
			j = 0;
			while(j != i) {
				k = sockwrite(outsock, xferdata, i - j);
				if(k < 0) {
					sockclose(outsock);
					insock = -1;
					return; /* let the caller close the file */
				}
				j = j + k;
			}
			if(hash != 0) {
				sizecount += i;
				while(sizecount > 127) {
					sizecount -= 128;
					putchar('#');
					fflush(stdout);
				}
			}
		}
	}
}

void docmd(VOIDFIX)
{
	int i, j, k;
	
	i = 0;
	j = strlen(cmdo);
	while(i != j) {
		k=sockwrite(sock, cmdo + i, j - i);
		if(k < 0) {
			doio(); /* this should kill the socket and program */
			exit(-2); /* just incase it doesn't */
		}
		i = i + k;
	}
}

/* buffered keystrokes */
void dokeyin(VOIDFIX)
{
	if(fixkbhit()) {
		key = fixgetch();
		if(depth < 126) {
			keyboard[depth] = key & 0xff;
			depth++;
			keyboard[depth] = 0x00;
		}
	}	
}

/* 
  wait for a non continuation message, printing all messages that apear,
  exit processing
  port passive processing
*/
void waitio(VOIDFIX)
{
	is = servstate;
	servstate = -1;
	
	while((servstate < 1)) {
		while(dav == 0) {
			doio();
			dokeyin();
		}
		printf("%s\n", smackit);
		if((smackit[3] != '-') && (strlen(smackit) > 2)) { /* acording to the spec, space is an option */
			servstate = atoi(smackit);
		}
		dav=0; /* clear data available on wait */
	}
	/* grab pasv here */
	
	if((servstate == 221) || (servstate == 421) || (servstate == 120)) {
		sockclose(sock);
		exit(0);
	} 
	if(servstate == 227) {
		/* scan till ( */
		while(smackit[0] != '(' && smackit[1] != 0x00) {
			memmove(smackit, smackit + 1, strlen(smackit) - 1);
		}
		memmove(smackit, smackit + 1, strlen(smackit) - 1);

		/* manually read in remote address and port (helps broken implementations) */
		
		/* sscanf(smackit, "%i,%i,%i,%i,%i,%i", h1, h2, h3, h4, p1, p2); */
		destination[0] = atoi(smackit);
		while(smackit[0] != ',' && smackit[1] != 0x00) {
			memmove(smackit, smackit + 1, strlen(smackit) - 1);
		}
		memmove(smackit, smackit + 1, strlen(smackit) - 1);
		destination[1] = atoi(smackit);
		while(smackit[0] != ',' && smackit[1] != 0x00) {
			memmove(smackit, smackit + 1, strlen(smackit) - 1);
		}
		memmove(smackit, smackit + 1,strlen(smackit) - 1);
		destination[2] = atoi(smackit);
		while(smackit[0] != ',' && smackit[1] != 0x00) {
			memmove(smackit, smackit + 1, strlen(smackit) - 1);
		}
		memmove(smackit, smackit + 1, strlen(smackit) - 1);
		destination[3] = atoi(smackit);
		while(smackit[0] != ',' && smackit[1] != 0x00) {
			memmove(smackit, smackit + 1, strlen(smackit) - 1);
		}
		memmove(smackit, smackit + 1, strlen(smackit) - 1);
		passport = 256 * atoi(smackit);
		while(smackit[0] != ',' && smackit[1] != 0x00) {
			memmove(smackit, smackit + 1, strlen(smackit) - 1);
		}
		memmove(smackit, smackit + 1, strlen(smackit) - 1);
		passport = passport + atoi(smackit);
	}
	
}

void doconin(VOIDFIX)
{
	register int i = 0;
	register int j = 0;
	int cr = 0;
	
	cmdq[0] = 0x00;
	if(depth > 0) {
		/* suck in all previous keystrokes */
		while((keyboard[j] != 0x0d) && (keyboard[j] != 0x00) && (keyboard[j] != 0x0a)) {
			if((keyboard[j] == 0x08) || (keyboard[j] == 0x7f)) {
				j++; /* backspace... */
				i--;
				cmdq[i] = 0x00;
			} else {
				cmdq[i] = keyboard[j];
				i++;
				j++;
			}
		}
		cmdq[i + 1] = 0x00;
		printf("%s", cmdq);
		fflush(stdout);
		if((keyboard[j] == 0x0d) || (keyboard[j] == 0x0a)) {
			cr = 1;
			j++;
		}
		depth = depth - j;
		memmove(keyboard, keyboard + j, depth);
		keyboard[depth] = 0x00;
	}
	while(cr == 0) {
		doio(); /* keep things going in the bg */
		if(fixkbhit()) {
			key = fixgetch();
			if((key == 0x08) || (key == 0x7f)) {
				if(i > 0) {
					cmdq[i] = 0x00;
					i--;
					printf("\b \b");
					fflush(stdout);
				}
			} else {
				if((key == 0x0d) || (key == 0x0a)) {
					cr = 1;
				} else {
					if(i < 126) {
						cmdq[i] = key & 0xff;
						i++;
						cmdq[i] = 0x00;
						putchar(key);
						fflush(stdout);
					}
				}
			}
		}
	}
	/* whew, we now have the data! */
	putchar('\n');
}

void doconinq(VOIDFIX)
{
	register int i = 0;
	register int j = 0;
	int cr = 0;
	
	cmdq[0] = 0x00;
	if(depth > 0) {
		/* suck in all previous keystrokes */
		while((keyboard[j] != 0x0d) && (keyboard[j] != 0x00) && (keyboard[j] != 0x0a)) {
			if((keyboard[j] == 0x08) || (keyboard[j] == 0x7f)) {
				j++; /* backspace... */
				i--;
				cmdq[i] = 0x00;
			} else {
				cmdq[i] = keyboard[j];
				i++;
				j++;
			}
		}
		cmdq[i+1] = 0x00;
		if(keyboard[j] == 0x0d) {
			cr = 1;
			j++;
		}
		depth = depth - j;
		memmove(keyboard, keyboard + j, depth);
		keyboard[depth] = 0x00;
	}
	while(cr == 0) {
		doio(); /* keep things going in the bg */
		if(fixkbhit()) {
			key = fixgetch();
			if((key == 0x08) || (key == 0x7f)) {
				if(i > 0) {
					cmdq[i] = 0x00;
					i--;
				}
			} else {
				if((key == 0x0d) || (key == 0x0a)) {
					cr = 1;
				} else {
					if(i < 126) {
						cmdq[i] = key & 0xff;
						i++;
						cmdq[i] = 0x00;
					}
				}
			}
		}
	}
	/* whew, we now have the data! */
	putchar('\n');
}


void dopasv(VOIDFIX)
{
	
	sprintf(cmdo, "PASV\r\n");
	docmd();
	servstate = -1;
}

void dohash(VOIDFIX)
{
	hash = 1 & (hash + 1); /* cheap xor for crap compiler :-( */
	printf("Hash printing now o");
	if(hash == 0) {
		printf("ff\n");
	} else {
		printf("n (128 bytes/hash mark)\n");
	}
	fflush(stdout);
	servstate = 200;
}

void dogetput(dir)
int dir;
{
	int i, j;
	sizecount = 0;
	sprintf(&cmdo[0], "TYPE I\r\n");
	if(dir == 2) sprintf(&cmdo[0], "TYPE A\r\n");
	docmd();
	servstate = -1;
	waitio();
	dopasv();
	waitio();
	if(servstate != 227) return;
	sprintf(cmdo, "RETR %s\r\n", cmdq);
	if(dir == 0) sprintf(cmdo, "STOR %s\r\n", cmdq);
	if(dir == 2) sprintf(cmdo, "LIST %s\r\n", cmdq);
	docmd();
	/* open passive connection */
	j = socket(AF_INET, SOCK_STREAM, 0);
	if(j < 0) {
		servstate = -1;
		return;
	}
	i = sockconnect(j, (int)passport, (unsigned char *) &destination);
	if(i < 0) {
		servstate = -1;
		return;
	}
	waitio();
	if(servstate != 150) {
		sockclose(j);
		return;
	}
	if(dir == 0) {
		outsock = j;
		if((xyz = fopen(cmdq, "rb")) == NULL) {
			printf("Can't open file %s\n", cmdq);
			xferfile = -1;
			/* close socket caught below... */
		} else {
			xferfile = fileno(xyz);
		}
		if(xferfile < 0) {
			sockclose(outsock);
			outsock = -1;
		}
	}
	if(dir == 1) {
		insock = j;
		if((xyz = fopen(cmdq, "wb")) == NULL) {
			printf("Can't create file %s\n",cmdq);
			xferfile = -1;
			/* close socket caught below... */
		} else {
			xferfile = fileno(xyz);
		}
		if(xferfile < 0) {
			sockclose(insock);
			insock = -1;
		}
	}
	if(dir == 2) {
		insock = j;
		xferfile = fileno(stdout); /* stdout */
		xyz = stdout;
	}
	while(insock > 0 || outsock > 0) {
		doio();
	}
	putchar('\n');
	if(dir == 0 || dir == 1) fclose(xyz); /* don't close stdout */
	servstate = -1;
	xyz = NULL;
}

void doparse(VOIDFIX)
{
	int i, l;
	/* 
	   chop off first part by changing space to a 0x00, 
	   after we know what/where the command is we memmove the data
	*/
	/* strip leading spaces */
	while(isspace(cmdq[0]) && (cmdq[0] != 0x00)) {
		memmove(cmdq, cmdq + 1, strlen(cmdq));
	}
	l = strlen(cmdq);
	if(l == 0) return; /* nothing to do */
	i = 0;
	while((!isspace(cmdq[i])) && (cmdq[i] != 0x00)) {
		i++;
	}
	if(cmdq[i] == 0x00) {
		cmdq[i + 1] = 0x00;
		l++;
	}
	cmdq[i] = 0x00; /* chop it */
	sprintf(cmdo, "%s", cmdq);
	i++;
	memmove(cmdq, cmdq + i,l - i + 1);
	/* strip leading spaces, again */
	while((isspace(cmdq[0])) && (cmdq[0] != 0x00)) {
		memmove(cmdq, cmdq + 1, strlen(cmdq));
	}
	l = strlen(cmdq);
	/* lower case typed command */
	for (i = 0; i < strlen(cmdo); i++) {
		cmdo[i] = tolower(cmdo[i]);
	}

	/* can't easily do this, it's going to be ugly, here goes */
	if((!strcmp(cmdo, "cd")) || (!strcmp(cmdo, "cwd"))) {
		if(l == 0) {
			printf("remote directory:");
			fflush(stdout);
			doconin();
		}
		sprintf(cmdo, "CWD %s\r\n", cmdq);
		docmd();
		servstate = -1;
	}

	if(!strcmp(cmdo, "pwd")) {
		sprintf(cmdo, "PWD\r\n", cmdq);
		docmd();
		servstate = -1;
	}

	if(!strcmp(cmdo, "put")) {
			dogetput(0);
	}

	if(!strcmp(cmdo, "get")) {
			dogetput(1);
	}

	if((!strcmp(cmdo, "ls")) || (!strcmp(cmdo, "dir"))) {
			if(l == 0) sprintf(cmdq, "-la"); /* unix */
			i = hash;
			hash = 0;
			dogetput(2);
			hash = i;
	}

	if((!strcmp(cmdo, "quit")) || (!strcmp(cmdo, "bye")) || (!strcmp(cmdo, "exit")) ) {
		sprintf(cmdo, "QUIT\r\n", cmdq);
		docmd();
		waitio(); /* get end message */
		waitio(); /* wait for socket to die off */
	}

	if(!strcmp(cmdo, "hash")) {
			dohash();
	}

}

void doprompt(VOIDFIX)
{
	printf("ftp>");
	fflush(stdout);
	doconin();
	doparse();
}

int main(argc, argv)
int argc;
char *argv[];
{
	struct u_nameip *result;
	int dotcount = 0;
	unsigned int port = 21;
	int xkb;
	char *netaddress;
	
	netaddress = malloc(80);
	keyboard = malloc(10);
	cmdq = malloc(160);
	xferdata = malloc(128);
	cmdo = malloc(160);
	iodata = malloc(BIGGEST);
	smackit = malloc(BIGGEST);

	if( (netaddress == (char *)NULL) || (keyboard == (char *)NULL) || (cmdq == (char *)NULL) || (xferdata == (char *)NULL) || (cmdo == (char *)NULL) || (iodata == (char *)NULL) || (smackit == (char *)NULL) ) {
		printf("Not enough memory\n");
		exit(1);
	}
	if(argc < 2 || argc > 3) AJKNETusage(PROGRAM, VERSION, EXECUTABLE, argv[0], "the.network.IP.address [port]");
	if(argc > 2) {
		sprintf(netaddress, "%s\n", argv[2]);
		port = atoi(netaddress) & 0xffff;
		if(port < 1 ) AJKNETusage(PROGRAM, VERSION, EXECUTABLE, argv[0], "the.network.IP.address [port]");
	}
/* IMPORTANT: initialize the network! */
	initnetwork(argc, argv);


	/* snag out the net address.... */
	result = findhost(argv[1],1);
	if(result == NULL) {
		printf("No such host\n");
	} else {

		/* take each one into a byte :-) */
		printf("\n\nTrying ");
		for(dotcount = 0; dotcount < 4; dotcount++) {
			printf("%i.", (short)(result->hostip[dotcount] & 0xff));
			destination[dotcount] = result->hostip[dotcount];
		}
		printf("\b (%s) port %i...", result->hostname, port);
		free(result->hostip);
		free(result->hostname);
		free(result);
	
		sock=socket(AF_INET, SOCK_STREAM, 0);
		if(sock == -1) {
			printf("No socket!\n");
		}
		if(sock > -1) zap = sockconnect(sock, (int)port, (unsigned char *) &destination);
	}
	if(zap > 0) {
		printf("\nConnected\n", sock,zap);
	} else {
		printf("\nService Not available\n");
	}

	xkb = 0;
	dotcount = 0;
	while(sock > 0) {
		 /* the macro `MAKEMEBLEED' should be sprinkled througout the app code
		 especially after a time consuming loop for latency reduction, like so: 

		( some c code, like a loop or something... )
		MAKEMEBLEED
		( more c code... )
		
		However if you are consistantly doing reads in a loop, this is done for you! :-)
		Writes _ALSO_ do this too, but that's because it's designed to and has to!
		A NULL write will cause the transport layer to hit once. SAME for a NULL read.
		THE SOCKET NEED NOT BE OPEN OR FUNCTIONING OR EVEN CORRECT.
		
		example:
		
		sockread(-1, NULL, 0); 
		is as valid as
		sockwrite(-1 ,NULL, 0);
		
		The file descriptor number does _NOT_ matter for it to work a POLL operation.
		The latency involved will be greater, though.
		
		The best way to make interactive programs work better is to read and process large chunks.
		Simple loops of reading 1 from socket will cause huge amounts of cpu load and latency.

		*/
		if(servstate < 0) waitio();
		if((servstate == 220) || (servstate == 530)) {
			printf("User Name:");
			fflush(stdout);
			doconin();
			sprintf(cmdo, "USER %s\r\n", cmdq);
			docmd();
			servstate = -1;
		} else if(servstate == 230) {
			dohash();
			sprintf(&cmdo[0], "SYST\r\n");
			docmd();
			servstate = -1;
		} else if(servstate == 331) {
			printf("Password:");
			fflush(stdout);
			doconinq();
			sprintf(cmdo, "PASS %s\r\n", cmdq);
			docmd();
			servstate = -1;
		} else if(servstate == 332) {
			printf("Account Password:");
			fflush(stdout);
			doconinq();
			sprintf(cmdo, "ACCT %s\r\n", cmdq);
			docmd();
			servstate = -1;
		} else if(servstate > 199) {
			doprompt();
		}
	}

	if(sock == 0) {
		printf("\n\nConection terminated\n");
	} else {
		printf("\n\nConection terminated with an error\n");
	}
	
#ifndef CPM
#ifndef z80
#ifndef HI_TECH_C
	exit(0);
#endif
#endif
#endif
}



/* tell link parser who we are and where we belong


0000 ftp.com ftp.obj xxPROGRAMxx

*/
